// Copyright (C) 2001 Jean-Marc Valin

#include <string>
#include <iostream>
#include "UINodeRepository.h"
#include "GUINodeParameters.h"
#include "GUINode.h"
#include "GUINetwork.h"
#include "Network.h"
#include "object_param.h"

using namespace std;

namespace FD {

static int user_count = 0;

static void param_apply(GnomePropertyBox *propertybox, gint arg1,
		GUINodeParameters* user_data) {
	//cerr << "Apply user_data = " << user_data << endl;
	user_data->apply();
}

static void param_close(GnomePropertyBox *propertybox,
		GUINodeParameters* user_data) {
	//cerr << "user_data = " << user_data << endl;
	user_data->hide();
}

static void entry_changed(GtkEntry *entry, GUINodeParameters* user_data) {
	user_data->changed();
}

static void type_changed(GnomePropertyBox *propertybox,
		GUINodeParameters* user_data) {
	user_data->changed();
}


/**
 *ダイアログのコンボボックスに変更があった場合に呼び出されるイベントハンドラ
 *
 */
static void value_changed(GtkMenu *optionmenu_menu, GUINodeParameters* user_data) {


	/// KEY POINT
	string changed_param = (char *)gtk_object_get_data(
			GTK_OBJECT(optionmenu_menu), "param_name");

	/// changed value
	GtkWidget *item_val = gtk_menu_get_active(GTK_MENU(optionmenu_menu));
	string changed_value =
			(char *)gtk_object_get_user_data(GTK_OBJECT(item_val));

	user_data->changed();

	/// 変更のあったwidgetを起点に、排他表示条件を再帰的に再設定する
	user_data->GUINodeParameters::setHideFlagRecurse(changed_param,
			changed_value, NULL, false, false);
	/// 表示しなおす
	user_data->hideFlagWidgetShowHide();

	user_data->show();

}

/** widget_visibleフラグを参照して パラメーターwidget を show or hide する
 *
 *
 */
void GUINodeParameters::hideFlagWidgetShowHide() {

	for (int i = 0; i < params.size(); i++) {
		if (params[i].widget_visible == false) {
			gtk_widget_hide(params[i].label); // LABEL
			gtk_widget_hide(params[i].optionmenu); // TYPE(OPTIONMENU)
			if (params[i].combo_entry) {
				gtk_widget_hide(params[i].combo_entry); // VALUE (GNOME COBO_ENTRY)
				gtk_widget_hide(params[i].entry); // VALUE (GNOME COBO_ENTRY)
			}
			if (params[i].optionmenu_val) {
				gtk_widget_hide(params[i].optionmenu_val); // VALUE (OPTIONMENU)
			}
		} else if (params[i].widget_visible == true) {
			gtk_widget_show(params[i].label); // LABEL
			gtk_widget_show(params[i].optionmenu); // TYPE(OPTIONMENU)
			if (params[i].combo_entry) {
				gtk_widget_show(params[i].combo_entry); // VALUE (GNOME COBO_ENTRY)
				gtk_widget_show(params[i].entry); // VALUE (GNOME COBO_ENTRY)
			}
			if (params[i].optionmenu_val) {
				gtk_widget_show(params[i].optionmenu_val); // VALUE (OPTIONMENU)
			}
		}
	}
	return;
}

/**指定されたパラメータを起点に、各パラメーターの不可視を再帰探索により決定していく
 * 不可視パラメータは、ParameterData::hyde_widgetである
 * parent_param	 : 起点パラメーター名
 * parent_value	 : 起点パラメータのVALUE
 * loop			 : 無限ループ防止集合
 * search_all	 : 現在の不可視状況を信用しないで探索する(true)
 * parent_hide	 : 起点パラメーターが表示されていない(true)
 *
 * 注意
 * 1.リストボックスの表示とレポジトリ(textParam)の内容に食い違いがある場合は、
 * textParamではなくて、widgetを参照している
 * 2. この探索アルゴリズムだと、探索回始点の parent_param の表示状態を決定できないが
 * 探索回始点の状態が変更されたのなら、可視であると言える。
 * 3. パラメーター名は、UINodeRepository, UINodeParameter のどちらから取得してもよい
 */
void GUINodeParameters::setHideFlagRecurse(string parent_param,
		string parent_value, set<string>* loop, bool search_all, bool force_hide) {

	NodeInfo *nodeInfo = UINodeRepository::Find(getUINode()->getType());

	/// 再帰レベル
	static int recurse_level = 0;
	recurse_level++;
/// cout << "-------------- HideFlagRecurse Lv. '" << recurse_level << "' : target_parent = '" << parent_param << "' value = '"<< parent_value << "' force_hide = '" << force_hide << endl;

	 /// 無限ループ防止フラグ(loop集合)確保
	if (loop == NULL || recurse_level == 1) {
		loop = new set<string>;
	}

	/// loop集合から、無限ループの判定をする
	set<string>::iterator it_loop = (*loop).begin();
	while (it_loop != (*loop).end()) {
		if ((*it_loop) == parent_param) { // Indirect looping
			cout << "Error: Found infinity recursive. target param :"
					<< parent_param << endl;
			return;
		}
		it_loop++;
	}
	///　判定が終わったら、loop集合に自身を追加する
	(*loop).insert(parent_param);

	for (int i = 0; i < nodeInfo->params.size(); i++) {
		map<string, ValidInfo*>::iterator it_valid_map = nodeInfo->params[i]->valid_map.begin();
///cout << "SEARCH_CHILD_LOOP " << i << endl;		
		/// 依存関係が在るか？
		it_valid_map = nodeInfo->params[i]->valid_map.find(parent_param);
		if (it_valid_map != nodeInfo->params[i]->valid_map.end()) {
			if (it_valid_map->second->valid_list.size() == 0) {
				continue;
			}
			ValidInfo *child_valid_info = it_valid_map->second;

			int child_index = i;
			// textParams[i]->name と  nodeInfo->params[i]->name は同一のはずである

			/// child valid_list から、依存関係において、不可視の状態を調査する
			bool found_target = false;
			for (int j = 0; j < child_valid_info->valid_list.size(); j++) {
				if (child_valid_info->valid_list[j] == parent_value) {
					found_target = true;
					break;
				}
			}

			/// 親が複数あることに備えて、不可視状態を親毎に保存する
			nodeInfo->params[child_index]->valid_map[parent_param]->visible_by_parent  = found_target;
			map<string, ValidInfo*>::iterator it = nodeInfo->params[child_index]->valid_map.begin();
///cout << "FOUND_TARGET : found_target=" << found_target << endl;
			bool multi_parent_visible = true;
			while (it != nodeInfo->params[child_index]->valid_map.end()) {
				multi_parent_visible &= it->second->visible_by_parent;
///cout << "MULTI_PARENT_VISIBLE : PARENT_PARAM=" << it->first << " visible=" << nodeInfo->params[child_index]->valid_map[parent_param]->visible_by_parent << endl;
				it++;
			}

			/// parent_value の値から、依存関係のある子のデータに状態変更が発生するかを change フラグとして記録する
			bool changed = false;
			bool next_force_hide = force_hide;
			if (force_hide == true) { /// 親が表示されないときは、依存している子は無条件に表示しない
				if (params[child_index].widget_visible == true) {
					changed = true;
				}
				params[child_index].widget_visible = false; // -> true(No Seeing)
				next_force_hide = true;
///cout << "SATAUS(" << textParams[child_index]->name << "): Set 'No Seeing(true)' (parent hide)"<< endl;
			} else {
				if (multi_parent_visible == true) {
					if (params[child_index].widget_visible == false) {
						changed = true;
					}
					params[child_index].widget_visible = true; // -> false(Seeing)
///cout << "SATAUS(" << textParams[child_index]->name << "): Set 'Seeing(false)'"<< endl;
				} else {
					if (params[child_index].widget_visible == true) {
						changed = true;
					}
					params[child_index].widget_visible = false; // -> true(No Seeing)
					next_force_hide = true;
///cout << "SATAUS(" << textParams[child_index]->name << "): Set 'No Seeing(false)'"<< endl;
				}
			}
			
			/// ここまで、状態変更の記録

			/// 状態変更が発生した場合は、次に依存するターゲットを再帰探索する
			/// search_allは、現在の状態が信用できない場合に用いる
			if (changed == true || search_all == true) {
///cout << "Parameter Changed. search_all =" << search_all << endl;

				/// 子パラメーターのVALUE、child_valueを取得する
				GtkWidget *widget_value = params[child_index].optionmenu_menu_val;
				string child_value;
				if (widget_value) {
					GtkWidget *item_val = gtk_menu_get_active(GTK_MENU(widget_value));
					child_value = (char *)gtk_object_get_user_data(GTK_OBJECT(item_val));
				} else {
					/// テキストエントリを親として、依存関係を持つことは在り得るが
					/// ここでは実装しない。
///cout << "Child parameter is TEXT ENTRY! child_index = " << child_index <<endl;
					continue;
					///		widget_value = gnome_entry_gtk_entry(GNOME_ENTRY(params[i].entry));
					///		if (!widget_value) {
					///			cout << "Error:  Can't get next target value." << endl;
					///		}
					child_value = (char *)gtk_entry_get_text(GTK_ENTRY(widget_value));
				}
///cout << "Call Recursive function" << endl;
				/// 再帰呼び出し
				setHideFlagRecurse(textParams[child_index]->name, child_value, loop, search_all, next_force_hide);

				/// loop 集合から自身を取り出す
				set<string>::iterator it_loop = (*loop).begin();
				while (it_loop != loop->end()) {
					if ((*it_loop) == nodeInfo->params[i]->name) {
///cout << "Erase Loop Set" << textParams[child_index]->name << endl;
						loop->erase(it_loop);
						break;
					}
					it_loop++;
				}
			}
		}
	}
	if (recurse_level == 1) {
		delete loop;
	}
	recurse_level--;
///cout << "-------------- HideFlagRecurse : Return " << endl;
	return;
}

/**
 *  全パラメーターを探索して、排他条件を、widget_visible フラグとして記録する
 *
 */
void GUINodeParameters::setHideFlagAllRecurse() {

	/// 初期の widget_visibleフラグは、全てtrueである。
	/// これは、setHideFlagRecurseの探索アルゴリズムが、トップノードの状態を判別できない為である
	/// ValidInfoのvisible_by_parentフラグの初期値がtrueなのは、複数の親に対してand検索を行う為
	
	for (int i = 0; i < params.size(); i++) {
///cout << "-------------- AllRecurse Start(" << i << ") --------------" << endl;
		if (params[i].optionmenu_menu_val == NULL) {
///cout << "optionmenu_menu_val is NULL. Loop count : " << i << endl;
			/// エントリーボックスが排他表示条件となる事はない
			continue;
		}
		GtkWidget *item_val =
				gtk_menu_get_active(GTK_MENU(params[i].optionmenu_menu_val));
		string active_value =
				(char *)gtk_object_get_user_data(GTK_OBJECT(item_val));
		setHideFlagRecurse(textParams[i]->name, active_value, NULL, true, false);
	}
///cout << "-------------- AllRecurse End --------------" << endl;
	return;
}


static void comments_changed(GtkTextView *entry, GUINodeParameters* user_data) {
	user_data->changed();
}

static void node_remove_input_select(GtkWidget *item,
		GUINodeParameters *node_param) {
	node_param->setInputSelect(item);
}

static void node_remove_input_unselect(GtkWidget *item,
		GUINodeParameters *node_param) {
	node_param->unsetInputSelect(item);
}

static void node_remove_output_select(GtkWidget *item,
		GUINodeParameters *node_param) {
	node_param->setOutputSelect(item);
}

static void node_remove_output_unselect(GtkWidget *item,
		GUINodeParameters *node_param) {
	node_param->unsetOutputSelect(item);
}

void GUINodeParameters::addInput() {
	char *new_input = (char*) gtk_entry_get_text(GTK_ENTRY(input_entry));
	if (!new_input || strlen(new_input)==0)
		return;

	gtk_list_select_all(GTK_LIST(list1));

	GtkWidget *inp1 = gtk_list_item_new_with_label(new_input);
	gtk_widget_ref(inp1);
	gtk_object_set_data(GTK_OBJECT(inp1), "label", (gpointer)new_input);
	gtk_signal_connect(GTK_OBJECT(inp1), "select",
			GTK_SIGNAL_FUNC(node_remove_input_select), this);
	gtk_signal_connect(GTK_OBJECT(inp1), "deselect",
			GTK_SIGNAL_FUNC(node_remove_input_unselect), this);
	gtk_object_set_data_full(GTK_OBJECT(list1), "table1", inp1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(inp1);
	GList *in_lis = g_list_append(NULL, inp1);

	gtk_list_append_items(GTK_LIST(list1), in_lis);
	node->addTerminal(string(new_input), UINetTerminal::INPUT);
}

void GUINodeParameters::addOutput() {

	char *new_output = (char*)gtk_entry_get_text(GTK_ENTRY(output_entry));

	if (!new_output || strlen(new_output)==0)
		return;

	gtk_list_select_all(GTK_LIST(list1));

	GtkWidget *outp1 = gtk_list_item_new_with_label(new_output);
	gtk_widget_ref(outp1);

	gtk_object_set_data(GTK_OBJECT(outp1), "label", (gpointer)new_output);
	gtk_signal_connect(GTK_OBJECT(outp1), "select",
			GTK_SIGNAL_FUNC(node_remove_output_select), this);
	gtk_signal_connect(GTK_OBJECT(outp1), "deselect",
			GTK_SIGNAL_FUNC(node_remove_output_unselect), this);
	gtk_object_set_data_full(GTK_OBJECT(list1), "table1", outp1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(outp1);
	GList *out_lis = g_list_append(NULL, outp1);

	gtk_list_append_items(GTK_LIST(list2), out_lis);
	node->addTerminal(string(new_output), UINetTerminal::OUTPUT);
}

void GUINodeParameters::removeInput() {
	if (!inputSelect) {
		//cerr << "no input selected\n";
		return;
	}
	char *input_name = (char *)gtk_object_get_data(GTK_OBJECT(inputSelect),
			"label");
	GList *lis = g_list_append(NULL, inputSelect);
	gtk_list_remove_items(GTK_LIST(list1), lis);
	node->removeTerminal(string(input_name), UINetTerminal::INPUT);
}

void GUINodeParameters::removeOutput() {
	if (!outputSelect) {
		//cerr << "no output selected\n";
		return;
	}
	char *output_name = (char *)gtk_object_get_data(GTK_OBJECT(outputSelect),
			"label");
	GList *lis = g_list_append(NULL, outputSelect);
	gtk_list_remove_items(GTK_LIST(list2), lis);
	node->removeTerminal(string(output_name), UINetTerminal::OUTPUT);
}

static void node_add_input(GtkButton *button, GUINodeParameters *node_param) {
	node_param->addInput();
}

static void node_add_output(GtkButton *button, GUINodeParameters *node_param) {
	node_param->addOutput();
}

static void node_remove_input(GtkButton *button, GUINodeParameters *node_param) {
	node_param->removeInput();
}

static void node_remove_output(GtkButton *button, GUINodeParameters *node_param) {
	node_param->removeOutput();
}

GUINodeParameters::GUINodeParameters(UINode *_node, string type) :
	UINodeParameters(_node, type), inputSelect(NULL), outputSelect(NULL),
			nodeproperty(NULL) {

}

GUINodeParameters::~GUINodeParameters() {
	if (nodeproperty) {
		gtk_widget_destroy(nodeproperty);
	}
}

void GUINodeParameters::show() {
	//create all widgets
	if (!nodeproperty) {
		createWindow();
	} else {
		gtk_widget_show(nodeproperty);
	}
}

void GUINodeParameters::hide() {
	//destroy widgets
	gtk_widget_destroy(nodeproperty);

	nodeproperty = NULL;

}

void GUINodeParameters::apply() {
	///Set this to true if there's any change involving a subnet_param change

	bool changedNetInterface=false;
	/// TYPE & VALUE
	for (int i=0; i<params.size(); i++) {
		/// VALUE
		string newValue;

		NodeInfo *nodeInfo = UINodeRepository::Find(getUINode()->getType());

		/// optionmenu 排他表示用リストメニューの場合
		if (nodeInfo != NULL && nodeInfo->params[i]->combo_list.size()) {
			GtkWidget *item_val = gtk_menu_get_active(GTK_MENU(params[i].optionmenu_menu_val));
			newValue = (char *)gtk_object_get_user_data(GTK_OBJECT(item_val));
		}
		/// Normal gnome entry　通常のテキストエントリーの場合
		else {
			///	GtkWidget *gtkentr = gnome_entry_gtk_entry(GNOME_ENTRY(params[i].entry));
			GtkWidget *gtkentr = params[i].combo_entry;
			newValue = gtk_entry_get_text(GTK_ENTRY(gtkentr));
		}
		/// TYPE
		string newType;
		///		GtkWidget *item =  params[i].optionmenu->menu_item;
		GtkWidget *item = gtk_menu_get_active(GTK_MENU(params[i].optionmenu_menu));
		newType = (char *)gtk_object_get_user_data(GTK_OBJECT(item));

		if (newType == "subnet_param" || textParams[i]->type == "subnet_param") {

			///There's a change somewhere
			if (newType != textParams[i]->type || newValue
					!= textParams[i]->value) {
				///subnet param
				changedNetInterface=true;
				///cerr<<"changedNetInterface = true"<<endl;
				///cerr<<"type : "<<newType<<endl;
				///cerr<<"value: "<<newValue<<endl;
			}
		}
		///Update type and value
		textParams[i]->type = newType;
		textParams[i]->value = newValue;
//	cout << "newType=" << newType << endl;
//	cout << "newValue=" << newValue << endl;

	}
	///nodeParams->setComments(string(gtk_editable_get_chars(GTK_EDITABLE(text_comments), 0, -1)));
	GtkTextBuffer *buffer =
			gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_comments));
	GtkTextIter start;
	GtkTextIter end;

	gtk_text_buffer_get_bounds(buffer, &start, &end);

	setComments(string(gtk_text_buffer_get_text(buffer, &start, &end, true)));

	node->getNetwork()->setModified();

	node->getNetwork()->interfaceChangeNotify();

}

void GUINodeParameters::changed() {
	gnome_property_box_changed(GNOME_PROPERTY_BOX(nodeproperty));
}

ParameterData *GUINodeParameters::getParamDataNamed(string n) {
	for (int i=0; i<textParams.size(); i++)
		if (textParams[i]->name == n)
			return &(params[i]);
	return NULL;
}

/*
 void GUINodeParameters::insertLoadedParam(ParameterText *param, string type, string value)
 {

 ParameterData *data = getParamDataNamed(param->name);
 const vector<string> &types=ObjectParam::allTypes();
 for (int i=0;i<types.size();i++)
 if (types[i] == type)
 gtk_option_menu_set_history (GTK_OPTION_MENU (data->optionmenu), i);
 GtkWidget *gtkentr = gnome_entry_gtk_entry(GNOME_ENTRY(data->entry));
 gtk_entry_set_text(GTK_ENTRY(gtkentr),(gchar *)value.c_str());
 }
 */

// ここから、パラメータポップアップウィンドウ
static void parameter_notify(GtkWidget *event_box, GdkEventMotion *event, GUINodeParameters* node_parameter){
	node_parameter->showParamPopup(event_box);
}
void GUINodeParameters::showParamPopup(GtkWidget *event_box) {

	GtkWidget *view;

	GdkColor color;
	gdk_color_parse ("#f0f055", &color); // 黄色

	// WINDOW
	GtkWidget *window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
	gtk_container_border_width(GTK_CONTAINER(window), 5);
	gtk_window_set_decorated(GTK_WINDOW (window), FALSE);
	gtk_widget_set_size_request (window, 600, -1);
	gtk_widget_modify_bg(window, GTK_STATE_NORMAL, &color);
	gtk_widget_show(window);

	// WINDOWポインタの保持
	param_popup_window = window;
	// テキストビューの作成
	view = gtk_text_view_new ();
	gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(view), GTK_WRAP_WORD);
	// フォント等
	PangoFontDescription *font_desc = pango_font_description_from_string ("Serif 13");
	gtk_widget_modify_font (view, font_desc);
	pango_font_description_free (font_desc);
	gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);
	gtk_widget_modify_base(view, GTK_STATE_NORMAL, &color);
	GtkTextBuffer *buffer = gtk_text_view_get_buffer (GTK_TEXT_VIEW (view));
	// 内容表示
#ifdef __LP64__
	long param_offset = (long)gtk_object_get_user_data(GTK_OBJECT(event_box));
#else
	int param_offset = (int)gtk_object_get_user_data(GTK_OBJECT(event_box));
#endif
	gtk_text_buffer_set_text (buffer, textParams[param_offset]->description.c_str(), -1);
	gtk_container_add(GTK_CONTAINER (window), view);
	gtk_widget_show (view);
}

void GUINodeParameters::deleteParamPopup() {
	gtk_widget_destroy(param_popup_window);
}
static void parameter_notify_delete(GtkWidget *widget, GdkEventMotion *event, GUINodeParameters* node_parameter) {
	node_parameter->deleteParamPopup();
}
GtkWidget *mygtk_event_box_new_attach(GtkWidget *attach, int param_offset, void *vp) {

	GtkWidget *event_box = gtk_event_box_new ();
	gtk_widget_show (event_box);
	gtk_container_add(GTK_CONTAINER (event_box), attach);
	gtk_widget_set_events (event_box, GDK_ENTER_NOTIFY_MASK|GDK_LEAVE_NOTIFY_MASK);
	gtk_object_set_user_data(GTK_OBJECT(event_box), (void *)param_offset);
	g_signal_connect (G_OBJECT (event_box), "enter_notify_event", G_CALLBACK(parameter_notify), vp);
	g_signal_connect (G_OBJECT (event_box), "leave_notify_event", G_CALLBACK(parameter_notify_delete), vp);
	return event_box;
}
// ここまで、パラメータポップアップウィンドウ

void GUINodeParameters::createWindow() {
	params.resize(textParams.size());
	//cerr << "GUINodeParameters::GUINodeParameters" << textParams.size() << "\n";

	int i;

	//GtkWidget *nodeproperty;
	GtkWidget *notebook2;
	GtkWidget *table2;
	GtkWidget *label14;
	GtkWidget *label15;
	GtkWidget *label16;
	GtkWidget *hseparator4;
	GtkWidget *hseparator5;
	GtkWidget *hseparator6;
	GtkWidget *glade_menuitem;

	GtkWidget *label12;
	GtkWidget *scrolledwindow2;
	GtkWidget *label13;
	GtkWidget *label_properties;

	GtkWidget *table1;
	GtkWidget *vseparator1;
	GtkWidget *vseparator2;
	GtkWidget *vseparator3;
	GtkWidget *hbox1;
	GtkWidget *button1;
	GtkWidget *button2;
	GtkWidget *hbox2;
	GtkWidget *button3;
	GtkWidget *button4;
	GtkWidget *label1;
	GtkWidget *label2;

	nodeproperty = gnome_property_box_new();

	//gnome_dialog_close_hides (GNOME_DIALOG(nodeproperty), TRUE);
	gtk_object_set_data(GTK_OBJECT(nodeproperty), "nodeproperty", nodeproperty);

	notebook2 = GNOME_PROPERTY_BOX (nodeproperty)->notebook;
	gtk_object_set_data(GTK_OBJECT(nodeproperty), "notebook2", notebook2);
	gtk_widget_show(notebook2);

	table2 = gtk_table_new(2+params.size(), 3, FALSE);
	gtk_widget_ref(table2);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "table2", table2,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(table2);

	gtk_container_add(GTK_CONTAINER(notebook2), table2);

	label14 = gtk_label_new(_("Name"));
	gtk_widget_ref(label14);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label14", label14,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label14);
	gtk_table_attach(GTK_TABLE(table2), label14, 0, 1, 0, 1,
			(GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);

	label15 = gtk_label_new(_("Type"));
	gtk_widget_ref(label15);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label15", label15,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label15);
	gtk_table_attach(GTK_TABLE(table2), label15, 1, 2, 0, 1,
			(GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);

	label16 = gtk_label_new(_("Value"));
	gtk_widget_ref(label16);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label16", label16,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label16);
	gtk_table_attach(GTK_TABLE(table2), label16, 2, 3, 0, 1,
			(GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);

	hseparator4 = gtk_hseparator_new();
	gtk_widget_ref(hseparator4);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "hseparator4",
			hseparator4, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(hseparator4);
	gtk_table_attach(GTK_TABLE(table2), hseparator4, 0, 1, 1, 2,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_EXPAND
					| GTK_FILL), 0, 0);

	hseparator5 = gtk_hseparator_new();
	gtk_widget_ref(hseparator5);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "hseparator5",
			hseparator5, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(hseparator5);
	gtk_table_attach(GTK_TABLE(table2), hseparator5, 1, 2, 1, 2,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);

	hseparator6 = gtk_hseparator_new();
	gtk_widget_ref(hseparator6);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "hseparator6",
			hseparator6, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(hseparator6);
	gtk_table_attach(GTK_TABLE(table2), hseparator6, 2, 3, 1, 2,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);

	NodeInfo *nodeInfo = UINodeRepository::Find(getUINode()->getType());
	for (i=0; i<params.size(); i++) {
		/// option menu of parameter name
		params[i].label = gtk_label_new(textParams[i]->name.c_str());
		gtk_widget_ref(params[i].label);
		gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label",
				params[i].label, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show(params[i].label);
		// descriptionが在る時だけポップアップのメニューを作る
		if (textParams[i]->description != "") {
			GtkWidget *event_box = mygtk_event_box_new_attach(params[i].label, i, this);
			gtk_table_attach(GTK_TABLE(table2), event_box, 0, 1, 2+i, 3+i,
					(GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);
		} else {
			gtk_table_attach(GTK_TABLE(table2), params[i].label, 0, 1, 2+i, 3+i,
					(GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);
		}
		/// option menu of parameter type
		params[i].optionmenu = gtk_option_menu_new();
		gtk_widget_ref(params[i].optionmenu);
		gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "optionmenu",
				params[i].optionmenu, (GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show(params[i].optionmenu);
		gtk_table_attach(GTK_TABLE(table2), params[i].optionmenu, 1, 2, 2+i, 3
				+i, (GtkAttachOptions) (0), (GtkAttachOptions) (0), 0, 0);

		/// Inserting parameter's type to optionmenu
		params[i].optionmenu_menu = gtk_menu_new();
		const vector<string> &types=ObjectParam::allTypes();
		for (int j=0; j<types.size(); j++) {
			glade_menuitem
					= gtk_menu_item_new_with_label((const gchar *)types[j].c_str());
			gtk_object_set_user_data(GTK_OBJECT(glade_menuitem),
					(void *)types[j].c_str());
			/// .def ファイルのパラメーターによって、TYPEのリストボックスを制限する
///			if (nodeInfo == NULL) { // SUBNETの時は.defファイル情報が無いので判定できない
				gtk_widget_show(glade_menuitem);
/*
			} else {
				if (nodeInfo->params[i]->type == "int"    ||
					nodeInfo->params[i]->type == "float"  ||
					nodeInfo->params[i]->type == "string" ||
					nodeInfo->params[i]->type == "bool")
				{
					if (types[j] == "subnet_param" && types[j] == nodeInfo->params[i]->type) {
						gtk_widget_show(glade_menuitem);
					}
				} else if (nodeInfo->params[i]->type == "any") {
					gtk_widget_show(glade_menuitem);
				} else {
///					type名にスペースなどが入っていて判別できない事も多い
///					if (types[j] == "object" || types[j] == "subnet_param") {
						gtk_widget_show(glade_menuitem);
///					}
				}
			}
*/
			gtk_menu_append(GTK_MENU(params[i].optionmenu_menu), glade_menuitem);
		}
		gtk_option_menu_set_menu(GTK_OPTION_MENU(params[i].optionmenu),
				params[i].optionmenu_menu);
		gtk_option_menu_set_history(GTK_OPTION_MENU(params[i].optionmenu), 0);

		gtk_signal_connect (GTK_OBJECT ( params[i].optionmenu_menu), "selection_done",
				GTK_SIGNAL_FUNC(type_changed), this);

		for (int k=0; k<types.size(); k++) {

			///cerr << types[k] << " vs " << textParams[i]->type << endl;
			if (types[k] == textParams[i]->type)
				gtk_option_menu_set_history(
						GTK_OPTION_MENU(params[i].optionmenu), k);
		}

		/// optionmenu of VALUE routine
		/// Count of parameter list
		NodeInfo *nodeInfo = UINodeRepository::Find(getUINode()->getType());

		if (nodeInfo && nodeInfo->params[i]->combo_list.size()) {
			params[i].entry = NULL; /// KEY POINT
			params[i].combo_entry = NULL; /// KEY POINT
			params[i].optionmenu_val = gtk_option_menu_new();

			gtk_widget_ref(params[i].optionmenu_val);
			gtk_object_set_data_full(GTK_OBJECT(nodeproperty),
					"optionmenu_val", params[i].optionmenu_val,
					(GtkDestroyNotify) gtk_widget_unref);
			gtk_widget_show(params[i].optionmenu_val);
			gtk_table_attach(GTK_TABLE(table2), params[i].optionmenu_val, 2, 3,
					2+i, 3 +i, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
					(GtkAttachOptions) (0), 0, 0);

			/// OPTION MENU MENU (VALUE) (pointer = optionmenu_menu_val)
			params[i].optionmenu_menu_val = gtk_menu_new();
			string cur_combo_list;
			for (int j = 0; j < nodeInfo->params[i]->combo_list.size(); j++) {
				cur_combo_list = nodeInfo->params[i]->combo_list[j];

				glade_menuitem = gtk_menu_item_new_with_label((const gchar *)(cur_combo_list.c_str()));
				gtk_object_set_user_data(GTK_OBJECT(glade_menuitem), (void *)(cur_combo_list.c_str()));
				gtk_widget_show(glade_menuitem);
				gtk_menu_append(GTK_MENU(params[i].optionmenu_menu_val),
						glade_menuitem);
			}
			gtk_option_menu_set_menu(GTK_OPTION_MENU(params[i].optionmenu_val),
					params[i].optionmenu_menu_val);
			gtk_option_menu_set_history(GTK_OPTION_MENU(params[i].optionmenu_val), 0);
			/// VALUEのリストボックス表示時のイベントハンドラ
			gtk_signal_connect(GTK_OBJECT(params[i].optionmenu_menu_val),
					"selection-done", GTK_SIGNAL_FUNC(value_changed), this);

			for (int j=0; j < nodeInfo->params[i]->combo_list.size(); j++) {
				if (nodeInfo->params[i]->combo_list[j] == textParams[i]->value) {
					gtk_option_menu_set_history(
							GTK_OPTION_MENU(params[i].optionmenu_val), j);
				}
			}
			/// Adding parameter's name to object data
			/// change_value シグナル時に使用するデータの埋め込み
			gtk_object_set_data(GTK_OBJECT(params[i].optionmenu_menu_val),
					"param_name", (void *)(textParams[i]->name.c_str()));

		/// Display normal EntryBox
		} else {
			/// GNOMEENTRY_OPTIONMENU
			params[i].optionmenu_val = NULL; /// KEYPOINT
			params[i].optionmenu_menu_val = NULL; /// KEYPOINT
			params[i].entry = gnome_entry_new(NULL);
			gtk_widget_ref(params[i].entry);
			gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "entry",
					params[i].entry, (GtkDestroyNotify) gtk_widget_unref);
			gtk_widget_show(params[i].entry);
			gtk_table_attach(GTK_TABLE(table2), params[i].entry, 2, 3, 2+i,
					3+i, (GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
					(GtkAttachOptions) (0), 0, 0);

			params[i].combo_entry
					= gnome_entry_gtk_entry(GNOME_ENTRY(params[i].entry));
			gtk_widget_ref(params[i].combo_entry);
			gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "combo_entry",
					params[i].combo_entry, (GtkDestroyNotify) gtk_widget_unref);
			gtk_widget_show(params[i].combo_entry);

			GtkWidget *gtkentr =
					gnome_entry_gtk_entry(GNOME_ENTRY(params[i].entry));
			gtk_signal_connect(GTK_OBJECT(gtkentr), "changed",
					GTK_SIGNAL_FUNC(entry_changed), this);

			gtk_entry_set_text(GTK_ENTRY(gtkentr),
					(gchar *)textParams[i]->value.c_str());
		}
	}

	label12 = gtk_label_new(_("Parameters"));
	gtk_widget_ref(label12);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label12", label12,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label12);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook2),
			gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook2), 0), label12);

	scrolledwindow2 = gtk_scrolled_window_new(NULL, NULL);
	gtk_widget_ref(scrolledwindow2);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "scrolledwindow2",
			scrolledwindow2, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(scrolledwindow2);
	gtk_container_add(GTK_CONTAINER(notebook2), scrolledwindow2);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolledwindow2),
			GTK_POLICY_NEVER, GTK_POLICY_ALWAYS);

	text_comments = gtk_text_view_new();
	gtk_widget_ref(text_comments);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "text_comments",
			text_comments, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(text_comments);
	gtk_container_add(GTK_CONTAINER(scrolledwindow2), text_comments);
	gtk_text_view_set_editable(GTK_TEXT_VIEW(text_comments), TRUE);

	label13 = gtk_label_new(_("Comments"));
	gtk_widget_ref(label13);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label13", label13,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label13);
	gtk_notebook_set_tab_label(GTK_NOTEBOOK(notebook2),
			gtk_notebook_get_nth_page(GTK_NOTEBOOK(notebook2), 1), label13);
	{
		int a=0;
		const char *str = getComments().c_str();
		gtk_text_buffer_set_text(
				gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_comments)), str, -1);

		//gtk_editable_insert_text (GTK_EDITABLE(text_comments), str, nodeParams->getComments().size(), &a);
	}
	g_signal_connect(
			G_OBJECT(gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_comments)) ),
			"changed", GTK_SIGNAL_FUNC(comments_changed), this);

	//properties tab (DL)
	//removed until it works


	GList *in_lis = NULL;
	GList *out_lis = NULL;

	vector<UITerminal *> inputs = node->getInputs();
	vector<UITerminal *> outputs = node->getOutputs();
	for (int i=0; i<inputs.size(); i++) {
		const char *input_name = inputs[i]->getName().c_str();
		GtkWidget *inp1 = gtk_list_item_new_with_label(input_name);
		gtk_widget_ref(inp1);
		gtk_object_set_data(GTK_OBJECT(inp1), "label", (gpointer)input_name);
		gtk_signal_connect(GTK_OBJECT(inp1), "select",
				GTK_SIGNAL_FUNC(node_remove_input_select), this);
		gtk_signal_connect(GTK_OBJECT(inp1), "deselect",
				GTK_SIGNAL_FUNC(node_remove_input_unselect), this);
		gtk_object_set_data_full(GTK_OBJECT(notebook2), "table1", inp1,
				(GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show(inp1);
		in_lis = g_list_append(in_lis, inp1);
	}

	for (int i=0; i<outputs.size(); i++) {
		const char *output_name = outputs[i]->getName().c_str();
		GtkWidget *outp1 = gtk_list_item_new_with_label(output_name);
		gtk_widget_ref(outp1);
		gtk_object_set_data(GTK_OBJECT(outp1), "label", (gpointer)output_name);
		gtk_signal_connect(GTK_OBJECT(outp1), "select",
				GTK_SIGNAL_FUNC(node_remove_output_select), this);
		gtk_signal_connect(GTK_OBJECT(outp1), "deselect",
				GTK_SIGNAL_FUNC(node_remove_output_unselect), this);
		gtk_object_set_data_full(GTK_OBJECT(notebook2), "table1", outp1,
				(GtkDestroyNotify) gtk_widget_unref);
		gtk_widget_show(outp1);
		out_lis = g_list_append(out_lis, outp1);
	}

	label_properties = gtk_label_new(_("Inputs/Outputs"));
	gtk_widget_ref(label_properties);
	gtk_object_set_data_full(GTK_OBJECT(nodeproperty), "label_properties",
			label_properties, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label_properties);

	//GtkWidget *notebook2;

	//window1 = gtk_window_new (GTK_WINDOW_TOPLEVEL);
	//gtk_object_set_data (GTK_OBJECT (window1), "window1", window1);
	//gtk_window_set_title (GTK_WINDOW (window1), _("window1"));

	table1 = gtk_table_new(3, 3, FALSE);
	gtk_widget_ref(table1);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "table1", table1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(table1);
	//gtk_container_add (GTK_CONTAINER (notebook2), table1);
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook2), table1, label_properties);

	vseparator1 = gtk_vseparator_new();
	gtk_widget_ref(vseparator1);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "vseparator1", vseparator1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(vseparator1);
	gtk_table_attach(GTK_TABLE(table1), vseparator1, 1, 2, 0, 1,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL), 0, 0);

	vseparator2 = gtk_vseparator_new();
	gtk_widget_ref(vseparator2);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "vseparator2", vseparator2,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(vseparator2);
	gtk_table_attach(GTK_TABLE(table1), vseparator2, 1, 2, 1, 2,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_EXPAND
					| GTK_FILL), 0, 0);

	vseparator3 = gtk_vseparator_new();
	gtk_widget_ref(vseparator3);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "vseparator3", vseparator3,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(vseparator3);
	gtk_table_attach(GTK_TABLE(table1), vseparator3, 1, 2, 2, 3,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_EXPAND
					| GTK_FILL), 0, 0);

	hbox1 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox1);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "hbox1", hbox1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(hbox1);
	gtk_table_attach(GTK_TABLE(table1), hbox1, 0, 1, 1, 2,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (GTK_FILL), 0, 0);

	input_entry = gtk_entry_new();
	gtk_widget_ref(input_entry);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "input_entry", input_entry,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(input_entry);
	gtk_box_pack_start(GTK_BOX(hbox1), input_entry, TRUE, TRUE, 0);

	button1 = gtk_button_new_with_label(_("Add"));
	gtk_widget_ref(button1);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "button1", button1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(button1);
	gtk_box_pack_start(GTK_BOX(hbox1), button1, FALSE, FALSE, 0);

	button2 = gtk_button_new_with_label(_("Remove"));
	gtk_widget_ref(button2);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "button2", button2,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(button2);
	gtk_box_pack_start(GTK_BOX(hbox1), button2, FALSE, FALSE, 0);

	hbox2 = gtk_hbox_new(FALSE, 0);
	gtk_widget_ref(hbox2);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "hbox2", hbox2,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(hbox2);
	gtk_table_attach(GTK_TABLE(table1), hbox2, 2, 3, 1, 2,
			(GtkAttachOptions) (GTK_EXPAND | GTK_FILL),
			(GtkAttachOptions) (GTK_FILL), 0, 0);

	output_entry = gtk_entry_new();
	gtk_widget_ref(output_entry);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "output_entry",
			output_entry, (GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(output_entry);
	gtk_box_pack_start(GTK_BOX(hbox2), output_entry, FALSE, FALSE, 0);

	button3 = gtk_button_new_with_label(_("Add"));
	gtk_widget_ref(button3);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "button3", button3,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(button3);
	gtk_box_pack_start(GTK_BOX(hbox2), button3, FALSE, FALSE, 0);

	button4 = gtk_button_new_with_label(_("Remove"));
	gtk_widget_ref(button4);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "button4", button4,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(button4);
	gtk_box_pack_start(GTK_BOX(hbox2), button4, FALSE, FALSE, 0);

	list2 = gtk_list_new();
	gtk_widget_ref(list2);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "list2", list2,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(list2);
	gtk_table_attach(GTK_TABLE(table1), list2, 2, 3, 2, 3,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);

	gtk_list_append_items(GTK_LIST(list2), out_lis);

	label1 = gtk_label_new(_("Inputs"));
	gtk_widget_ref(label1);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "label1", label1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label1);
	gtk_table_attach(GTK_TABLE(table1), label1, 0, 1, 0, 1,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment(GTK_MISC(label1), 0, 0.5);

	label2 = gtk_label_new(_("Outputs"));
	gtk_widget_ref(label2);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "label2", label2,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(label2);
	gtk_table_attach(GTK_TABLE(table1), label2, 2, 3, 0, 1,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (0), 0, 0);
	gtk_misc_set_alignment(GTK_MISC(label2), 0, 0.5);

	list1 = gtk_list_new();
	gtk_widget_ref(list1);
	gtk_object_set_data_full(GTK_OBJECT(notebook2), "list1", list1,
			(GtkDestroyNotify) gtk_widget_unref);
	gtk_widget_show(list1);
	gtk_table_attach(GTK_TABLE(table1), list1, 0, 1, 2, 3,
			(GtkAttachOptions) (GTK_FILL), (GtkAttachOptions) (GTK_FILL), 0, 0);

	gtk_list_append_items(GTK_LIST(list1), in_lis);

	gtk_signal_connect(GTK_OBJECT(button1), "clicked",
			GTK_SIGNAL_FUNC(node_add_input), this);
	gtk_signal_connect(GTK_OBJECT(button2), "clicked",
			GTK_SIGNAL_FUNC(node_remove_input), this);
	gtk_signal_connect(GTK_OBJECT(button3), "clicked",
			GTK_SIGNAL_FUNC(node_add_output), this);
	gtk_signal_connect(GTK_OBJECT(button4), "clicked",
			GTK_SIGNAL_FUNC(node_remove_output), this);

	//gnome_property_box_changed(GNOME_PROPERTY_BOX(nodeproperty));
	gtk_signal_connect(GTK_OBJECT(nodeproperty), "apply",
			GTK_SIGNAL_FUNC(param_apply), this);

	gtk_signal_connect(GTK_OBJECT(nodeproperty), "close",
			GTK_SIGNAL_FUNC(param_close), this);

	gnome_property_box_set_state(GNOME_PROPERTY_BOX(nodeproperty), false);

	gtk_widget_show(nodeproperty);

	/// パラメーターの全依存関係を探索して、表示フラグを設定する
	/// 全依存関係を探索するのは最初だけである、２回目以降は変更があった物だけでよい
	setHideFlagAllRecurse();
	/// フラグに応じてwidget_show を行う
	hideFlagWidgetShowHide();

	//cerr << "this is " << this << "\n";

}

}//namespace FD
